# MEGA TODO https://docs.python.org/3/library/multiprocessing.shared_memory.html --> can I block shared memory to prevent tf issues ???
# https://stackoverflow.com/questions/14124588/shared-memory-in-multiprocessing --> or maybe use the real true shared mem ???
# https://stackoverflow.com/questions/55277138/tensorflow-keras-multiprocessing-results-in-lack-of-gpu-memory --> mayeb this is my solution
# https://stackoverflow.com/questions/43641543/how-to-avoid-loading-a-parent-module-in-a-forked-process-with-pythons-multiproce
# https://stackoverflow.com/questions/42504669/keras-tensorflow-and-multiprocessing-in-python --> ok this is how I should do that then maybe --> TODO give it a try --> quite easy TODO

# https://github.com/yuanyuanli85/Keras-Multiple-Process-Prediction

# https://stackoverflow.com/questions/70263918/how-to-prevent-multiprocessing-from-inheriting-imports-and-globals --> maybe I need spawn --> can I get it to run ??? and if that does not work can I get it to default to something else
# can I close deep learning and tensorflow before MT processing --> maybe ???
# to reprodcue the error --> run deep learning first then run ta finish all and everything will be stuck forever

# do a small test to reproduce the MT error --> probably not so hard in fact
